package com.persagy.bdtp.adm.service;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.persagy.bdtp.adm.common.AdmResult;
import com.persagy.bdtp.adm.entity.Pagination;
import com.persagy.bdtp.adm.entity.oldadm.OldAdmResult;
import com.persagy.dmp.basic.model.QueryCriteria;
import com.persagy.dmp.common.model.response.CommonResult;

import java.util.function.Supplier;

public class ServiceUtil {

	private static final String SUCCESS = "success";

	public static <T> T call(Supplier<CommonResult<T>> action) {
		CommonResult<T> result = action.get();
		if(SUCCESS.equals(result.getResult()))
			return result.getData();
		throw new CallException(new AdmResult<>(result.getResult(), result.getMessage()));
	}

	public static <T> T callOldAdm(Supplier<OldAdmResult<T>> action, Pagination pagination) {
		OldAdmResult<T> result = action.get();
		if(SUCCESS.equals(result.getResult())) {
			if(pagination != null)
				pagination.setTotal(result.getTotal());
			return result.getContent();
		}
		throw new CallException(new AdmResult<>(result.getResult(), result.getMessage()));
	}

	public static <T> T callOldAdm(Supplier<OldAdmResult<T>> action) {
		return callOldAdm(action, null);
	}

	public static QueryCriteria getQueryCriteria(String objType, ObjectMapper objectMapper){
		QueryCriteria queryCriteria = new QueryCriteria();

		ObjectNode objectNode = objectMapper.createObjectNode();
		if(objType != null)
			objectNode.put("objType", objType);

		queryCriteria.setCriteria(objectNode);
		return queryCriteria;
	}

}
