package com.persagy.bdtp.adm.service.impl;

import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.persagy.bdtp.adm.client.DicClient;
import com.persagy.bdtp.adm.client.RwdClient;
import com.persagy.bdtp.adm.common.AdmConst;
import com.persagy.bdtp.adm.common.AdmResult;
import com.persagy.bdtp.adm.dao.*;
import com.persagy.bdtp.adm.entity.*;
import com.persagy.bdtp.adm.entity.oldadm.OProject;
import com.persagy.bdtp.adm.service.*;
import com.persagy.dmp.basic.model.QueryCriteria;
import com.persagy.dmp.common.constant.ResponseCode;
import com.persagy.dmp.common.helper.SpringHelper;
import com.persagy.dmp.define.entity.RelationDefine;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.*;

@Service
public class SyncAppImpl implements ISyncApp {

	@Autowired
	private IAdmObjectService objectService;

	@Autowired
	private IQueryFromOldAdm queryOldAdm;

	@Autowired
	private RwdClient rwdClient;

	@Autowired
	private DicClient dicClient;

	@Autowired
	private IConfigService configService;

	@Autowired
	private AdmJobSpaceMapper jobSpaceMapper;

	@Autowired
	private AdmProblemArchMapper problemArchMapper;

	@Autowired
	private AdmProblemEquipMapper problemEquipMapper;

	@Autowired
	private AdmRelationMapper relationMapper;

	@Autowired
	private AdmFileMapper fileMapper;

	@Autowired
	private AdmQrCodeMapper qrCodeMapper;

	@Override
	public Map<String, Object> downloadDictAndConfig(String groupCode, String projectId, String userId) {
		Map<String, Object> dict = queryDict(groupCode, projectId, userId);

		//TODO 配置数据查询
		List<AdmM2dEquip> m2dTypes = configService.queryM2dEquip(projectId);
		List<AdmInfosConfig> infosConfig = configService.queryInfosConfig(projectId);
		List<AdmRelsConfig> relConfig = configService.queryRelsConfig(projectId);
		List<AdmPipeConfig> pipeConfig = configService.queryPipeConfig(projectId);
		List<AdmConfig> config = configService.queryCommonConfig(projectId);
		List<AdmComponent> component = configService.queryComponent(projectId);

		HashMap<String, Object> data = new HashMap<>();
		data.put("dict", dict);
		data.put("m2dTypes", m2dTypes);
		data.put("infosConfig", infosConfig);
		data.put("relConfig", relConfig);
		data.put("pipeConfig", pipeConfig);
		data.put("config", config);
		data.put("component", component);

		return data;
	}

	private Map<String, Object> queryDict(String groupCode, String projectId, String userId){
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);

		//从知识库服务获取专业分类
		QueryCriteria majorCriteria = ServiceUtil.getQueryCriteria(null, objectMapper);
		majorCriteria.getCriteria().put("type", "major");
		List<Map<String, Object>> majorList = ServiceUtil.call (() -> dicClient.dataQuery(groupCode, projectId, AdmConst.APP_ID, userId, majorCriteria));

		//查询系统和设备定义
		QueryCriteria criteria = ServiceUtil.getQueryCriteria(null, objectMapper);
		criteria.getCriteria().putArray("objType").add("system").add("equipment");
		List<TypeDef> typeList = ServiceUtil.call(() -> rwdClient.queryObjectType(groupCode, projectId, AdmConst.APP_ID, userId, criteria));

		//查询关系定义
		List<RelationDefine> relList = ServiceUtil.call(() -> rwdClient.queryRelDef(groupCode, projectId, AdmConst.APP_ID, userId, new QueryCriteria()));

		LinkedList<TypeDef> sysList = new LinkedList<>();
		LinkedList<TypeDef> equipList = new LinkedList<>();
		for(TypeDef def : typeList){
			if(def.getObjType().equals("system"))
				sysList.add(def);
			else if(def.getObjType().equals("equipment"))
				equipList.add(def);
		}

		//组织树形结构
//		for(Map<String, Object> major : majorList){
//			ArrayList<TypeDef> childrenSys = new ArrayList<>();
//			major.put("children", childrenSys);
//			String majorCode = (String) major.get("code");
//			for(Iterator<TypeDef> sysIter = sysList.iterator();sysIter.hasNext();){
//				TypeDef sysDef = sysIter.next();
//				if(sysDef.getMajorCode() != null && sysDef.getMajorCode().equals(majorCode)){
//					childrenSys.add(sysDef);
//
//					for(Iterator<TypeDef> equipIter = equipList.iterator();equipIter.hasNext();){
//						TypeDef equipDef = equipIter.next();
//						if(equipDef.getSystemCode() != null && equipDef.getSystemCode().equals(sysDef.getCode())){
//							sysDef.addChild(equipDef);
//							equipIter.remove();
//						}
//					}
//
//					sysIter.remove();
//					break;
//				}
//			}
//		}

		HashMap<String, Object> data = new HashMap<>();
		data.put("major", majorList);
		data.put("system", sysList);
		data.put("equipment", equipList);
		data.put("relation", relList);
		return data;
	}

	@Override
	public Map<String, Object> downloadFrameData(String groupCode, String projectId, String userId) {
		List<OProject> prjList = null;
		String token = queryOldAdm.login();
		if(StrUtil.isNotBlank(token))
			prjList = queryOldAdm.queryProject(token, groupCode, projectId);
		else
			throw new CallException(new AdmResult(ResponseCode.A0220.getCode(), "ADM platform: " + ResponseCode.A0220.getDesc()));

		String[] prjIds = new String[prjList.size()];
		for (int i = 0; i < prjIds.length; i++) {
			prjIds[i] = prjList.get(i).getId();
		}
		List<AdmObject> bdAndFl = objectService.queryBdAndFl(prjIds);

		HashMap<String, Object> data = new HashMap<>();
		data.put("projects", prjList);
		data.put("buildingsAndFloors", bdAndFl);
		return data;
	}

	@Override
	public Map<String, Object> downloadFloorData(String projectId, String floorId) {
		List<AdmObject> floorData = objectService.queryByFloor(projectId, floorId);
		List<AdmRelation> floorRel = relationMapper.selectList(buildQueryWrapperForFloor(new QueryWrapper<AdmRelation>(), projectId, floorId, true));

		//TODO 任务相关数据查询
		List<AdmJobSpace> jobSpace = jobSpaceMapper.selectList(buildQueryWrapperForFloor(new QueryWrapper<AdmJobSpace>(), projectId, floorId, false));
		List<AdmProblemArch> problemArch = problemArchMapper.selectList(buildQueryWrapperForFloor(new QueryWrapper<AdmProblemArch>(), projectId, floorId, false));
		List<AdmProblemEquip> problemEquip = problemEquipMapper.selectList(buildQueryWrapperForFloor(new QueryWrapper<AdmProblemEquip>(), projectId, floorId, false));
		List<AdmFile> file = fileMapper.selectList(buildQueryWrapperForFloor(new QueryWrapper<AdmFile>(), projectId, floorId, false));
		List<AdmQrCode> qrCode = qrCodeMapper.selectList(buildQueryWrapperForFloor(new QueryWrapper<AdmQrCode>(), projectId, floorId, false));

		HashMap<String, Object> data = new HashMap<>();
		data.put("floorId", floorId);

		data.put("objects", floorData);
		data.put("relations", floorRel);

		data.put("jobSpace", jobSpace);
		data.put("problemArch", problemArch);
		data.put("problemEquip", problemEquip);
		data.put("file", file);
		data.put("qrCode", qrCode);

		return data;
	}

	private <T> QueryWrapper<T> buildQueryWrapperForFloor(QueryWrapper<T> wrapper, String projectId, String floorId, boolean floorIdNull){
		wrapper.eq("project_id", projectId);
		if(!floorIdNull)
			wrapper.eq("floor_id", floorId);
		else
			wrapper.and(w -> w.eq("floor_id", floorId).or().isNull("floor_id"));
		return wrapper;
	}

}
