package com.persagy.bdtp.adm.util;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.persagy.bdtp.adm.common.AdmMessageConst;
import com.persagy.dmp.amqp.handler.RabbitMessageSender;
import com.persagy.dmp.common.exception.BusinessException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
@Slf4j
public class AdmMessageHelper {

    @Autowired
    private RabbitMessageSender messageSender;

    @Autowired
    private ObjectMapper objectMapper;

    public void sendMessage(Object messageData, String routing, boolean syncFlag) {
        // 同步消息
        try {
            String msg = objectMapper.writeValueAsString(messageData);
            if(syncFlag) {
                messageSender.sendAndReceive(AdmMessageConst.ADM_EXCHANGE, routing, msg);
            } else {
                // 异步消息
                messageSender.send(AdmMessageConst.ADM_EXCHANGE, routing, msg);
            }
        } catch (JsonProcessingException e) {
            log.error(e.getMessage(), e);
            throw new BusinessException("消息格式有误！");
        }
    }

}
